package common

import (
	"crypto/md5"
	"fmt"
	"log"
	"sync"
	"time"
)

type Metric struct {
	Name   string
	Labels [][]string
	Value  float64
	// Optional accumulator functionality
	Accumulator *Accumulator
	Type        string
	TTL         time.Duration
}

type Measures map[string]*Metric

type MetricData struct {
	sync.RWMutex
	Measures Measures
}

type accValue struct {
	value float64
	ts    time.Time
}

type Accumulator struct {
	sync.Mutex
	ttl        time.Duration
	values     []*accValue
	counter    uint64
	outputType string
}

func (acc *Accumulator) Values() (res map[string]float64) {
	acc.clean()

	values := acc.values
	if len(values) == 0 {
		// FIXME
		// Setting a value of 0 can be wrong for the max/mix
		// And IS wrong for the average
		values = []*accValue{{value: 0}}
	}

	operations := []string{"sum"}
	if acc.outputType == "stats" {
		operations = []string{"max", "avg", "min"}
	}

	res = map[string]float64{
		"total": float64(acc.counter),
	}
	max, min, sum := acc.Compute(values)
	for _, op := range operations {
		switch op {
		case "sum":
			res["sum"] = sum
		case "max":
			res["max"] = max
		case "avg":
			res["avg"] = sum / float64(len(values))
		case "min":
			res["min"] = min
		}
	}
	return
}

func (acc *Accumulator) Compute(values []*accValue) (max, min, sum float64) {
	min = float64(1e20)
	for _, value := range values {
		v := value.value
		sum += v
		if v < min {
			min = v
		}
		if v > max {
			max = v
		}
	}
	return
}

func (acc *Accumulator) Push(value float64) {
	acc.values = append(acc.values, &accValue{
		ts:    time.Now(),
		value: value,
	})
	acc.counter++
}

// Cleanup values older than the TTL
func (acc *Accumulator) clean() {
	acc.Lock()
	defer acc.Unlock()

	values := acc.values
	if len(values) == 0 {
		return
	}

	for i, v := range acc.values {
		if time.Since(v.ts) < acc.ttl {
			acc.values = values[i:]

			return
		}
	}
}

func MakeMetricData() *MetricData {
	return &MetricData{
		Measures: make(map[string]*Metric),
	}
}

func (m *MetricData) GetMetricsByName(metricName string) []*Metric {
	m.Lock()
	metrics := make([]*Metric, 0)
	for _, metric := range m.Measures {
		if metric.Name == metricName {
			metrics = append(metrics, metric)
		}
	}
	m.Unlock()

	return metrics
}

func (m *MetricData) IncrError(module, serviceType string) {
	m.Acc(&Metric{
		Name: "self_errors",
		Labels: [][]string{
			[]string{"module", module},
			[]string{"service_type", serviceType},
		},
		Value: 1,
		Type:  "counter",
		TTL:   time.Duration(10) * time.Second,
	})
}

func (m *MetricData) fingerprint(metric Metric) string {
	payload := metric.Name
	for _, l := range metric.Labels {
		payload += l[1]
	}
	return fmt.Sprintf("%x", md5.Sum([]byte(payload)))
}

func (m *MetricData) Set(metric *Metric) {
	m.Lock()
	fp := m.fingerprint(*metric)
	m.Measures[fp] = metric
	m.Unlock()
}

func (m *MetricData) Acc(metric *Metric) {
	if metric.TTL == time.Duration(0) || metric.Type == "" {
		log.Println("WARN", "in accumulator", "missing TTL/Type init values", metric)
		return
	}
	m.Lock()
	fp := m.fingerprint(*metric)
	if _, ok := m.Measures[fp]; !ok {
		metric.Accumulator = &Accumulator{
			ttl:        metric.TTL,
			outputType: metric.Type,
			values:     []*accValue{},
		}
		m.Measures[fp] = metric
	}
	m.Measures[fp].Accumulator.Push(metric.Value)
	m.Unlock()
}

func (m *Metric) GetLabel(searchLabel string) (string, bool) {
	for _, label := range m.Labels {
		if label[0] == searchLabel {
			return label[1], true
		}
	}
	return "", false
}
