package healthcheck

import (
	"context"
	"encoding/base64"
	"fmt"
	"io/ioutil"
	"log"
	"net"
	"net/http"
	"os"
	"regexp"
	"strings"
	"sync"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

type hcDaemon struct {
	daemon.Daemon

	targets           []*common.Service
	defaultHealchecks string
	healthcheckPath   string

	healthchecks map[string]healthcheck
}

type module struct {
	Name        string            `yaml:"name"`
	Prober      string            `yaml:"prober"`
	TimeoutStr  string            `yaml:"timeout"`
	Send        string            `yaml:"send"`
	ExpectStr   string            `yaml:"expect"`
	Method      string            `yaml:"method"`
	Headers     map[string]string `yaml:"headers"`
	StatusCodes []int             `yaml:"status_codes"`
	Route       string            `yaml:"route"`
	Timeout     time.Duration     `yaml:"_disabled"`
	Expect      *regexp.Regexp    `yaml:"_disabled2"`
}

type healthcheck interface {
	check(common.Service) (bool, error)
}

// HTTPCheck Defines a generic HTTP check
type HTTPCheck struct {
	*module
}

func (hc *HTTPCheck) check(target common.Service) (passed bool, err error) {
	req, err := http.NewRequest(hc.Method, "http://"+target.IP+fmt.Sprintf(":%d", target.Port)+hc.Route, strings.NewReader(hc.Send))
	if err != nil {
		return false, err
	}
	ctx, cancel := context.WithTimeout(context.Background(), hc.Timeout)
	defer cancel()

	resp, err := http.DefaultClient.Do(req.WithContext(ctx))
	if err != nil {
		return false, err
	}
	defer resp.Body.Close()
	// Check for valid status codes
	passed = false
	for _, code := range hc.StatusCodes {
		if code == resp.StatusCode {
			passed = true
			break
		}
	}
	if !passed {
		return passed, nil
	}
	if hc.Expect == nil {
		return true, nil
	}
	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return false, err
	}
	return hc.Expect.Match(body), nil
}

// TCPCheck Defines a generic TCP check
type TCPCheck struct {
	*module
}

func (hc *TCPCheck) check(target common.Service) (passed bool, err error) {
	conn, err := net.DialTimeout("tcp", fmt.Sprintf("%s:%d", target.IP, target.Port), hc.Timeout)
	if err != nil {
		return false, err
	}
	defer conn.Close()

	if err := conn.SetDeadline(time.Now().Add(hc.Timeout)); err != nil {
		return false, err
	}

	if hc.Send == "" && hc.Expect == nil {
		return true, nil
	}

	if _, err = conn.Write([]byte(hc.Send)); err != nil {
		return false, err
	}
	var res = make([]byte, 1e6)

	if _, err := conn.Read(res); err != nil {
		return false, err
	}
	return hc.Expect.Match(res), nil
}

func (d *hcDaemon) parseHealthchecks(path string, checkStr string) (map[string]healthcheck, error) {
	healthchecks := make(map[string]healthcheck)

	var hc struct {
		Modules []*module
	}

	if path == "" {
		data, err := base64.StdEncoding.DecodeString(checkStr)
		if err != nil {
			return healthchecks, err
		}

		if err := yaml.Unmarshal(data, &hc); err != nil {
			return healthchecks, err
		}
		log.Println("Healthchecks definition loaded from defaults")
	} else {
		f, err := os.Open(path)
		if err != nil {
			return healthchecks, err
		}
		defer f.Close()

		if err := yaml.NewDecoder(f).Decode(&hc); err != nil {
			return healthchecks, err
		}
		log.Printf("Healthchecks definition loaded from %s", path)
	}

	for _, module := range hc.Modules {
		var err error

		if module.Prober == "none" {
			continue
		}

		ok := true
		if module.Timeout, err = time.ParseDuration(module.TimeoutStr); err != nil {
			log.Printf(`WARN failed to parse 'timeout' for module %s: "%s"`, module.Name, err)
			ok = false
		}
		if module.ExpectStr != "" {
			if module.Expect, err = regexp.Compile(module.ExpectStr); err != nil {
				log.Printf(`WARN failed to parse 'expect' for module %s: "%s"`, module.Name, err)
				ok = false
			}
		}
		switch module.Prober {
		case "http":
			healthchecks[module.Name] = &HTTPCheck{module}
		case "tcp":
			healthchecks[module.Name] = &TCPCheck{module}
		default:
			log.Printf(`WARN unknown 'prober' for module %s: "%s"`, module.Name, module.Prober)
			ok = false
		}

		if !ok {
			log.Printf("WARN ignoring module %s", module.Name)
			continue
		}
	}
	return healthchecks, nil
}

func MakeDaemon(hc string, healthcheckPath string) *hcDaemon {
	return &hcDaemon{
		defaultHealchecks: hc,
		healthcheckPath:   healthcheckPath,
	}
}

func (d *hcDaemon) ID() string {
	return "Healthchecks"
}

func (d *hcDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	return d.reload(targets)
}

func (d *hcDaemon) OnTargetsChange(targets []*common.Service) error {
	return d.reload(targets)
}

func (d *hcDaemon) reload(targets []*common.Service) error {
	healthchecks, err := d.parseHealthchecks(d.healthcheckPath, d.defaultHealchecks)
	if err != nil {
		return fmt.Errorf("healthchecks loading failed: %w", err)
	}

	log.Printf("Loaded %d healthchecks", len(healthchecks))

	enabledTargets := make([]*common.Service, 0, len(targets))
	for _, target := range targets {
		if _, ok := healthchecks[target.Type]; !ok {
			log.Printf("WARN unknown healthcheck %s for service %s", target.Type, target.ID)

			continue
		}

		enabledTargets = append(enabledTargets, target)
	}

	d.healthchecks = healthchecks
	d.targets = enabledTargets

	return nil
}

func (d *hcDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup
	for _, target := range d.targets {
		hc := d.healthchecks[target.Type]

		wg.Add(1)
		go func(target *common.Service) {
			defer wg.Done()

			val := float64(0)
			passed, _ := hc.check(*target)
			if passed {
				val = float64(1)
			}
			m.Set(&common.Metric{
				Name: "probe_success",
				Labels: [][]string{
					[]string{"service", target.ID},
					[]string{"service_type", target.Type},
					[]string{"addr", fmt.Sprintf("%s:%d", target.IP, target.Port)},
				},
				Value: val,
			})
		}(target)
	}

	wg.Wait()

	return nil
}
