package log

import (
	"context"
	"encoding/base64"
	"fmt"
	"log"
	"os"
	"regexp"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/hpcloud/tail"
	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

type logDaemon struct {
	daemon.Daemon

	services        []*common.Service
	logPaths        map[string]string
	ttl             time.Duration
	defaultPatterns string
	patternPath     string
	patterns        map[string]*regexp.Regexp
	rawPatterns     map[string]string

	wg    sync.WaitGroup
	stopC chan struct{}
}

var tailConfig = tail.Config{
	// Logger: tail.DiscardingLogger,
	Follow: true,
	// This prevents channel closure on file moves
	ReOpen: true,
	// This prevents tail from re-parsing previous lines
	Location: &tail.SeekInfo{Offset: 0, Whence: 2},
}

var servicesWithLogFile = []string{"oioswift"}

func parsePatterns(path, defaultPatterns string) (map[string]*regexp.Regexp, map[string]string, error) {
	rawPatterns := make(map[string]string)
	if path == "" {
		data, err := base64.StdEncoding.DecodeString(defaultPatterns)
		if err != nil {
			return nil, nil, fmt.Errorf("Failed to base64-decode default patterns: %w", err)
		}

		if err := yaml.Unmarshal(data, &rawPatterns); err != nil {
			return nil, nil, fmt.Errorf("Failed to yaml-decode default patterns: %w", err)
		}

		log.Println("Log patterns definition loaded from defaults")
	} else {
		f, err := os.Open(path)
		if err != nil {
			return nil, nil, fmt.Errorf("Failed to open patterns file: %w", err)
		}
		defer f.Close()

		if err := yaml.NewDecoder(f).Decode(&rawPatterns); err != nil {
			return nil, nil, fmt.Errorf("Failed to decode patterns: %w", err)
		}
		log.Printf("Log patterns definition loaded from %s", path)
	}

	compiledPatterns := make(map[string]*regexp.Regexp)
	for serviceType, rawPattern := range rawPatterns {
		re, err := regexp.Compile(rawPattern)
		if err != nil {
			return nil, nil, fmt.Errorf("Failed to compile pattern `%s` for service '%s': %w", rawPattern, serviceType, err)
		}

		compiledPatterns[serviceType] = re
	}

	return compiledPatterns, rawPatterns, nil
}

func MakeDaemon(defaultPatterns, patternPath string, ttl time.Duration) *logDaemon {
	return &logDaemon{
		ttl:             ttl,
		defaultPatterns: defaultPatterns,
		patternPath:     patternPath,
	}
}

func (d *logDaemon) ID() string {
	return "Logs"
}

func (d *logDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	return d.reload(services)
}

func (d *logDaemon) OnServicesChange(services common.ServicesMap) error {
	return d.reload(services)
}

func (d *logDaemon) reload(services common.ServicesMap) error {
	logPfx := "/var/log/oio/sds/"

	compiledPatterns, rawPatterns, err := parsePatterns(d.patternPath, d.defaultPatterns)
	if err != nil {
		return fmt.Errorf("Failed to load patterns: %w", err)
	}
	log.Printf("INFO: loaded %d log patterns", len(compiledPatterns))
	d.patterns = compiledPatterns
	d.rawPatterns = rawPatterns

	enabledServices := make([]*common.Service, 0, 2*len(services))
	logPaths := make(map[string]string)
	for svcType, instances := range services {
		if _, ok := compiledPatterns[svcType]; !ok {
			log.Println("WARN", "Service type", svcType, "has no configured pattern")
			continue
		}
		for _, svc := range instances {
			if logPath, ok := svc.Conf["log_location"]; ok {
				enabledServices = append(enabledServices, svc)
				logPaths[svc.ID] = logPath
				continue
			}

			// TODO: handle service types with "-" in their names
			// TODO: handle "subservices" (0.1, 0.2 ...)
			parts := strings.Split(svc.ID, "-")
			if len(parts) != 3 {
				log.Println("WARN", "Could not autoparse log location for svc", svc)
				continue
			}
			sfx := ".access"
			if common.Index(servicesWithLogFile, svcType) > -1 {
				sfx = ".log"
			}
			if svcType == "rawx" {
				sfx = "-httpd-access.log"
			}
			enabledServices = append(enabledServices, svc)
			logPaths[svc.ID] = fmt.Sprintf("%s%s/%s-%s%s",
				logPfx,
				strings.Replace(svc.ID, "-", "/", 1),
				parts[1],
				parts[2],
				sfx)
		}
	}

	if d.stopC != nil {
		close(d.stopC)
		d.wg.Wait()
	}

	d.services = enabledServices
	d.logPaths = logPaths
	d.stopC = nil

	return nil
}

func (d *logDaemon) Collect(m *common.MetricData) error {
	if d.stopC != nil {
		return nil
	}

	d.wg = sync.WaitGroup{}
	d.stopC = make(chan struct{})

	for _, svc := range d.services {
		logPath := d.logPaths[svc.ID]

		d.wg.Add(1)
		go func(svc *common.Service) {
			defer d.wg.Done()

			d.tailLogFile(svc, logPath, m)
		}(svc)
	}

	return nil
}

// TODO: If there is an error before we manage to watch the file, retry during the next Collect
func (d *logDaemon) tailLogFile(svc *common.Service, logPath string, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	rTokenize := d.patterns[svc.Type]

	// TODO: Add this in the patterns file as options
	// detect dot in response time
	multiplier := float64(1)
	for _, v := range strings.Split(d.rawPatterns[svc.Type], " ") {
		if strings.Contains(v, "resp_time") {
			if strings.Contains(v, ".") {
				multiplier = 1000
			}
			break
		}
	}

	numTokens := len(rTokenize.SubexpNames())

	baseLabels := [][]string{
		[]string{"addr", target},
		[]string{"service", svc.ID},
		[]string{"service_type", svc.Type},
	}

	var t *tail.Tail
	for {
		var err error
		if t, err = tailFile(logPath); err == nil {
			break
		}

		m.IncrError("log", svc.Type)
		log.Printf("WARN: Could not watch file '%s': %s. Retrying in %s", logPath, err, d.ttl)

		select {
		case <-d.stopC:
			return
		case <-time.After(d.ttl):
		}
	}

	defer t.Cleanup()
	defer func() {
		if err := t.Stop(); err != nil {
			log.Printf("WARN: Error stopping tailing file '%s': %s", logPath, err)
		}
	}()

	var line *tail.Line
	for {
		// Same trick as in the daemon package
		select {
		case <-d.stopC:
			return
		default:
		}

		select {
		case line = <-t.Lines:
		case <-d.stopC:
			return
		}

		if line.Err != nil {
			log.Printf("WARN Error getting line: %s", line.Err)
			continue
		}

		if err := d.handleLine(line.Text, rTokenize, numTokens, multiplier, baseLabels, m); err != nil {
			m.IncrError("log", svc.Type)
			log.Printf("WARN Error handling line from file '%s': %s", logPath, err)
		}
	}
}

func tailFile(filePath string) (*tail.Tail, error) {
	if _, err := os.Stat(filePath); err != nil {
		return nil, err
	}
	return tail.TailFile(filePath, tailConfig)
}

func (d *logDaemon) handleLine(line string, rTokenize *regexp.Regexp, numTokens int, multiplier float64, baseLabels [][]string, m *common.MetricData) error {
	s := rTokenize.FindStringSubmatch(line)
	if len(s) != numTokens {
		// TODO: add a debug line to see when a line is skipped
		return nil
	}
	names := rTokenize.SubexpNames()[1:]

	baseLabels = append(baseLabels, []string{"method", s[1+common.Index(names, "method")]})
	for idx, tag := range names {
		if strings.HasPrefix(tag, "tag_") {
			baseLabels = append(baseLabels, []string{strings.Split(tag, "tag_")[1], s[idx+1]})
		}
	}
	for idx, name := range names {
		if common.Index([]string{"bytes_sent", "resp_length", "resp_time", "code"}, name) == -1 {
			continue
		}

		value := s[idx+1]
		if value == "" || value == "-" {
			continue
		}

		vFloat, err := strconv.ParseFloat(value, 4)
		if err != nil {
			return fmt.Errorf("Could not convert '%s' to float: %w", name, err)
		}

		baseLabelsLen := len(baseLabels)
		baseLabelsCopy := make([][]string, baseLabelsLen, baseLabelsLen+1)
		copy(baseLabelsCopy, baseLabels)

		switch name {
		case "bytes_sent":
			m.Acc(&common.Metric{
				Name:   "log_bandwidth",
				Labels: append(baseLabelsCopy, []string{"type", "in"}),
				Value:  vFloat,
				Type:   "stats",
				TTL:    d.ttl,
			})
		case "resp_length":
			m.Acc(&common.Metric{
				Name:   "log_bandwidth",
				Labels: append(baseLabelsCopy, []string{"type", "out"}),
				Value:  vFloat,
				Type:   "stats",
				TTL:    d.ttl,
			})
		case "resp_time":
			m.Acc(&common.Metric{
				Name:   "log_response_time",
				Labels: baseLabelsCopy,
				Value:  vFloat * multiplier,
				Type:   "stats",
				TTL:    d.ttl,
			})
		case "code":
			m.Acc(&common.Metric{
				Name:   "log_response_code",
				Labels: append(baseLabelsCopy, []string{"code", s[common.Index(names, "code")+1]}),
				Value:  1,
				Type:   "counter",
				TTL:    d.ttl,
			})
		}
	}

	return nil
}

func (d *logDaemon) Close(ctx context.Context) error {
	if d.stopC == nil {
		return nil
	}

	close(d.stopC)

	done := make(chan struct{})
	go func() {
		d.wg.Wait()
		close(done)
	}()

	select {
	case <-done:
		return nil
	case <-ctx.Done():
		return ctx.Err()
	}
}
