package main

import (
	"context"
	"errors"
	"flag"
	"fmt"
	"io"
	"log"
	"net/http"
	"os"
	"os/signal"
	"strings"
	"sync"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
	hc "github.com/open-io/oio-exporter/healthcheck"
	self "github.com/open-io/oio-exporter/internal"
	clog "github.com/open-io/oio-exporter/log"
	"github.com/open-io/oio-exporter/oiofs"
	"github.com/open-io/oio-exporter/openio"
	"github.com/open-io/oio-exporter/redis"
	"github.com/open-io/oio-exporter/s3rt"
	"github.com/open-io/oio-exporter/static"
	"github.com/open-io/oio-exporter/tcp"
	"github.com/open-io/oio-exporter/version"
	"gopkg.in/yaml.v2"
)

var HEALTHCHECKS = ""
var LOGPATTERNS = ""
var VERSIONS = ""

type inventory struct {
	Namespaces map[string]struct {
		Services  map[string][]*common.Service `yaml:"services"`
		Localname string                       `yaml:"localname"`
		Rack      string                       `yaml:"rack"`
	} `yaml:"namespaces"`
}

type oioExporter struct {
	host      string
	bind      string
	inventory string
	ns        string
	localname string
	rack      string
	hcTargets []string
	metrics   *common.MetricData
	prefix    string

	state *common.State
}

// Export prometheus metrics
func (exp *oioExporter) export(w http.ResponseWriter, r *http.Request) {
	filter := r.URL.Query().Get("filter")
	exp.metrics.RLock()
	for _, metric := range exp.metrics.Measures {
		if filter != "" && !strings.HasPrefix(metric.Name, filter) {
			continue
		}

		labelStr := []string{exp.host}
		for _, labelPair := range metric.Labels {
			labelStr = append(labelStr, labelPair[0]+"=\""+labelPair[1]+"\"")
		}

		if metric.Accumulator != nil {
			for k, v := range metric.Accumulator.Values() {
				fmt.Fprintf(w, "%s%s{%s} %f\n", exp.prefix, metric.Name+"_"+k, strings.Join(labelStr, ","), v)
			}
		} else {
			fmt.Fprintf(w, "%s%s{%s} %f\n", exp.prefix, metric.Name, strings.Join(labelStr, ","), metric.Value)
		}
	}
	exp.metrics.RUnlock()
}

// Reload local service configuration
func (exp *oioExporter) reload(w http.ResponseWriter, r *http.Request) {
	n, err := exp.loadLocalConfig()
	if err != nil {
		log.Fatalln(err)
	}
	log.Println("Reloaded", n, "local targets")
	fmt.Fprintf(w, "OK")
}

// Serve local service configuration
func (exp *oioExporter) config(w http.ResponseWriter, r *http.Request) {
	file, err := os.Open(exp.inventory)
	if err != nil {
		log.Fatal(err)
	}
	if _, err := io.Copy(w, file); err != nil {
		log.Println("WARN", "could not serve config", err)
	}
}

func (exp *oioExporter) parseConfig(r io.Reader) (*inventory, error) {
	var inv *inventory
	if err := yaml.NewDecoder(r).Decode(&inv); err != nil {
		return nil, err
	}

	if inv == nil || len(inv.Namespaces) == 0 {
		return nil, errors.New("Empty inventory")
	}

	for _, servicesMap := range inv.Namespaces {
		for serviceType, services := range servicesMap.Services {
			for _, service := range services {
				service.Type = serviceType
			}
		}
	}

	return inv, nil
}

func (exp *oioExporter) loadLocalConfig() (int, error) {
	file, err := os.Open(exp.inventory)
	if err != nil {
		return 0, err
	}
	defer file.Close()

	inv, err := exp.parseConfig(file)
	if err != nil {
		return 0, err
	}

	var namespace string
	for ns := range inv.Namespaces {
		namespace = ns
		break
	}

	exp.ns = namespace
	exp.localname = inv.Namespaces[namespace].Localname
	exp.rack = inv.Namespaces[namespace].Rack
	n := len(inv.Namespaces[namespace].Services)

	exp.state.OnServicesChange.L.Lock()
	exp.state.ServicesLastUpdate = time.Now()
	exp.state.Services = inv.Namespaces[namespace].Services
	exp.state.OnServicesChange.Broadcast()
	exp.state.OnServicesChange.L.Unlock()

	return n, nil
}

func (exp *oioExporter) loadRemoteConfig(target string) (int, error) {
	services := []*common.Service{}
	res, err := http.Get("http://" + target + "/config")
	if err != nil {
		return 0, err
	}
	defer res.Body.Close()

	inv, err := exp.parseConfig(res.Body)
	if err != nil {
		return 0, err
	}

	var remoteServices common.ServicesMap
	for _, services := range inv.Namespaces {
		remoteServices = services.Services
		break
	}
	for svcType, instances := range remoteServices {
		for _, svc := range instances {
			svc.Type = svcType
			services = append(services, svc)
		}
	}

	n := len(services)

	exp.state.OnTargetsChange.L.Lock()
	exp.state.TargetsLastUpdate = time.Now()
	exp.state.Targets = append(exp.state.Targets, services...)
	exp.state.OnTargetsChange.Broadcast()
	exp.state.OnTargetsChange.L.Unlock()

	return n, nil
}

func (exp *oioExporter) reloadRemote() {
	retry := 10 * time.Second

	var wg sync.WaitGroup
	wg.Add(len(exp.hcTargets))

	for _, target := range exp.hcTargets {
		go func(target string) {
			defer wg.Done()
			for {
				n, err := exp.loadRemoteConfig(target)
				if err != nil {
					log.Printf("WARN Error loading remote config '%s': %s; retrying in %s", target, err, retry)
					time.Sleep(retry)
					continue
				}

				log.Println("Loaded", n, "remote targets")

				return
			}
		}(target)
	}

	wg.Wait()

	log.Print("INFO Loaded all remote targets")
}

func (exp *oioExporter) listen(stop <-chan struct{}) {
	http.HandleFunc("/metrics", exp.export)
	http.HandleFunc("/config", exp.config)
	http.HandleFunc("/reload", exp.reload)
	server := &http.Server{Addr: exp.bind}

	go func() {
		if err := server.ListenAndServe(); err != nil && !errors.Is(err, http.ErrServerClosed) {
			log.Print(err)
		}
	}()

	<-stop

	if err := server.Shutdown(context.TODO()); err != nil {
		log.Print(err)
	}
}

func main() {
	var inventory string
	var targets string
	var patternPath string
	var healthcheckPath string
	var versionPath string
	var bind string
	var daemonsToRun string

	host, err := os.Hostname()
	if err != nil {
		log.Println("WARN", "Could not retrieve hostname", err)
	}

	fs := flag.NewFlagSet("", flag.ExitOnError)
	fs.StringVar(&inventory, "inventory", "/etc/oio/sds/OPENIO/inventory.yml", "Path to inventory file")
	fs.StringVar(&patternPath, "pattern-path", "", "Path to log patterns")
	fs.StringVar(&healthcheckPath, "healthcheck-path", "", "Path to healthchecks definition")
	fs.StringVar(&versionPath, "version-path", "", "Path to versions definition")
	fs.StringVar(&host, "host", host, "Report metrics with the specified hostname")
	fs.StringVar(&bind, "bind", "0.0.0.0:9115", "Listen address")
	fs.StringVar(&targets, "targets", "localhost:9115", "Comma separated list of targets to perform healthchecks on")
	fs.StringVar(&daemonsToRun, "daemons", "", "Comme separated list of daemons to run (empty means all of them)")
	if err := fs.Parse(os.Args[1:]); err != nil {
		log.Fatalln("Could not parse args", err)
	}

	exp := oioExporter{
		host:      fmt.Sprintf(`host="%s"`, host),
		bind:      bind,
		inventory: inventory,
		hcTargets: strings.Split(targets, ","),
		metrics:   common.MakeMetricData(),
		prefix:    "openio_",
		state:     common.NewState(),
	}

	n, err := exp.loadLocalConfig()
	if err != nil {
		log.Fatalln("FATAL", "in loading local config", err)
	}
	log.Println("Loaded", n, "local targets")

	go exp.reloadRemote()

	allDaemonNames := "hc,oio,redis,tcp,version,s3rt,log,oiofs,self,static"
	if len(daemonsToRun) == 0 {
		daemonsToRun = allDaemonNames
	}

	daemonNames := strings.Split(daemonsToRun, ",")
	daemons := make([]daemon.DaemonI, len(daemonNames))
	for i, daemonName := range daemonNames {
		switch daemonName {
		case "hc":
			daemons[i] = hc.MakeDaemon(HEALTHCHECKS, healthcheckPath)
		case "oio":
			daemons[i] = openio.MakeDaemon(exp.ns)
		case "redis":
			daemons[i] = redis.MakeDaemon()
		case "tcp":
			daemons[i] = tcp.MakeDaemon()
		case "version":
			daemons[i] = version.MakeDaemon(VERSIONS, versionPath)
		case "s3rt":
			daemons[i] = s3rt.MakeDaemon()
		case "log":
			daemons[i] = clog.MakeDaemon(LOGPATTERNS, patternPath, 10*time.Second)
		case "oiofs":
			daemons[i] = oiofs.MakeDaemon()
		case "self":
			daemons[i] = self.MakeDaemon()
		case "static":
			daemons[i] = static.MakeDaemon(exp.ns, exp.localname, exp.rack)
		default:
			log.Fatalf("ERROR Unknown daemon %s (daemons are: %s)", daemonName, allDaemonNames)
		}
	}

	sigCh := make(chan os.Signal, 1)
	signal.Notify(sigCh, os.Interrupt)

	stop := make(chan struct{})
	var wg sync.WaitGroup
	wg.Add(1 + len(daemons))

	go func() {
		defer wg.Done()
		exp.listen(stop)
	}()

	for _, d := range daemons {
		go func(d daemon.DaemonI) {
			defer wg.Done()

			log.Printf("INFO Start %s", d.ID())
			if err := daemon.Run(d, exp.metrics, exp.state, stop); err != nil {
				// TODO: match Init errors with errors.Is & retry, return only for Close errors
				log.Printf("ERROR %s: %s", d.ID(), err)

				return
			}
			log.Printf("INFO Closed %s", d.ID())
		}(d)
	}

	<-sigCh

	log.Print("Closing gracefully")

	close(stop)

	// Free the reloading goroutines
	exp.state.OnServicesChange.L.Lock()
	exp.state.OnServicesChange.Broadcast()
	exp.state.OnServicesChange.L.Unlock()
	exp.state.OnTargetsChange.L.Lock()
	exp.state.OnTargetsChange.Broadcast()
	exp.state.OnTargetsChange.L.Unlock()

	waitCh := make(chan struct{})
	go func() {
		defer close(waitCh)
		wg.Wait()
	}()

	select {
	case <-waitCh:
		log.Print("Exited")
	case <-time.After(20 * time.Second):
		log.Fatal("Timeout")
	case <-sigCh:
		log.Fatal("Forced Exit")
	}
}
