package openio

import (
	"fmt"
	"log"
	"os"
	"strconv"
	"strings"
	"sync"
	"syscall"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
)

type oioDaemon struct {
	daemon.Daemon

	proxyURL string
	ns       string
	types    []string
	services common.ServicesMap
}

type metaxInfoBody struct {
	Cache     map[string]int64
	Elections map[string]int64
}

type serviceInfoBody struct {
	Addr  string
	Score int
}

func MakeDaemon(ns string) *oioDaemon {
	return &oioDaemon{
		ns: ns,
	}
}

func (d *oioDaemon) ID() string {
	return "OIO"
}

func (d *oioDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	return d.reload(services)
}

func (d *oioDaemon) OnServicesChange(services common.ServicesMap) error {
	return d.reload(services)
}

func (d *oioDaemon) reload(services common.ServicesMap) error {
	proxyURL := ""
	if svc, ok := services["oioproxy"]; ok && len(svc) > 0 {
		proxyURL = fmt.Sprintf("http://%s:%d/v3.0/", svc[0].IP, svc[0].Port)
	}

	var types []string
	url := proxyURL + d.ns + "/conscience/info?what=types"
	if err := common.HTTPGetJson(url, &types); err != nil {
		return fmt.Errorf("failed to fetch service types from conscience: %w", err)
	}

	d.services = services
	d.proxyURL = proxyURL
	d.types = types

	return nil
}

func (d *oioDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup
	for serviceType, services := range d.services {
		if common.Index(d.types, serviceType) == -1 {
			continue
		}

		wg.Add(1)
		go func(serviceType string, services []*common.Service) {
			defer wg.Done()

			d.collectScore(serviceType, services, m)
		}(serviceType, services)

		for _, svc := range services {
			wg.Add(2)
			go func(service *common.Service) {
				defer wg.Done()

				d.collectVolumeInfo(*service, m)
			}(svc)
			go func(serviceType string, service *common.Service) {
				defer wg.Done()

				switch serviceType {
				case "meta0", "meta1", "meta2":
					d.collectMeta(*service, m)
				case "rawx":
					d.collectRawx(*service, m)
				}
			}(serviceType, svc)
		}
	}

	wg.Wait()

	return nil
}

func (d *oioDaemon) collectMeta(svc common.Service, m *common.MetricData) {
	if d.proxyURL == "" {
		return
	}
	// TODO: move target as service address
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	counters, err := common.HTTPGetCsv(d.proxyURL+"forward/stats?id="+target, ' ')
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN: metaX stats collection failed", err)
		return
	}
	for _, cntr := range counters {
		val, err := strconv.ParseFloat(cntr[2], 64)
		if err != nil {
			continue
		}
		if cntr[0] == "counter" {
			parts := strings.Split(cntr[1], ".")
			if len(parts) != 3 {
				continue
			}
			m.Set(&common.Metric{
				Name: parts[0] + "_" + parts[1],
				Labels: [][]string{
					[]string{"dimension", parts[2]},
					[]string{"service", svc.ID},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
					// TODO: add volume
				},
				Value: val,
			})
		}
	}
	svcInfo := metaxInfoBody{}
	err = common.HTTPGetJson(d.proxyURL+"forward/info?id="+target, &svcInfo)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting meta2 info", svc.IP, svc.Port, err)
		return
	}
	for t, metrics := range map[string]map[string]int64{"cache": svcInfo.Cache, "elections": svcInfo.Elections} {
		for dim, val := range metrics {
			m.Set(&common.Metric{
				Name: t,
				Labels: [][]string{
					[]string{"dimension", dim},
					[]string{"service", svc.ID},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
				},
				Value: float64(val),
			})
		}
	}
}

func (d *oioDaemon) collectRawx(svc common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	counters, err := common.HTTPGetCsv("http://"+target+"/stat", ' ')
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN: rawx metric collection failed", err)
		return
	}
	for _, cntr := range counters {
		val, err := strconv.ParseFloat(cntr[2], 64)
		if err != nil {
			continue
		}
		if cntr[0] == "counter" {
			parts := strings.Split(cntr[1], ".")
			if len(parts) != 3 {
				continue
			}
			m.Set(&common.Metric{
				Name: parts[0] + "_" + parts[1],
				Labels: [][]string{
					[]string{"dimension", parts[2]},
					[]string{"service", svc.ID},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
					// TODO: add volume
				},
				Value: val,
			})
		}
	}
}

func (d *oioDaemon) collectScore(svcType string, services []*common.Service, m *common.MetricData) {
	info := []serviceInfoBody{}
	err := common.HTTPGetJson(d.proxyURL+d.ns+"/conscience/list?type="+svcType, &info)
	if err != nil {
		m.IncrError("openio", svcType)
		log.Println("WARN", "when collecting score info", svcType, err)
		return
	}
	for _, svc := range services {
		target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
		for _, remoteSvc := range info {
			if remoteSvc.Addr == target {
				m.Set(&common.Metric{
					Name: "score",
					Labels: [][]string{
						[]string{"service", svc.ID},
						[]string{"service_type", svc.Type},
						[]string{"addr", target},
						// TODO: add volume
					},
					Value: float64(remoteSvc.Score),
				})
				break
			}
		}
	}
}

func (d *oioDaemon) collectVolumeInfo(svc common.Service, m *common.MetricData) {
	if svc.Volume == "" {
		return
	}
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)
	f, err := os.OpenFile(svc.Volume, os.O_RDONLY, 0644)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting volume info", svc, err)
		return
	}
	defer f.Close()

	var stat syscall.Statfs_t
	err = syscall.Fstatfs(int(f.Fd()), &stat)
	if err != nil {
		m.IncrError("openio", svc.Type)
		log.Println("WARN", "when collecting volume info", svc, err)
		return
	}
	vMetric := map[string]uint64{
		"byte_avail":  stat.Bavail * uint64(stat.Bsize),
		"byte_used":   (stat.Blocks - stat.Bfree) * uint64(stat.Bsize),
		"byte_free":   stat.Bfree * uint64(stat.Bsize),
		"inodes_free": stat.Ffree,
		"inodes_used": stat.Files - stat.Ffree,
	}
	fsID := strconv.FormatUint(uint64(stat.Fsid.X__val[1])<<32|uint64(stat.Fsid.X__val[0]), 10)
	for dim, val := range vMetric {
		m.Set(&common.Metric{
			Name: dim,
			Labels: [][]string{
				[]string{"service", svc.ID},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
				[]string{"volume", svc.Volume},
				[]string{"fsid", fsID},
			},
			Value: float64(val),
		})
	}
}
