package redis

import (
	"context"
	"encoding/json"
	"fmt"
	"log"
	"regexp"
	"strconv"
	"strings"
	"sync"

	"github.com/go-redis/redis"
	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
)

var scriptGetAccounts = redis.NewScript(`
    return redis.call("hgetall", "accounts:")
`)

var scriptGetContCount = redis.NewScript(`
    local key = "containers:" .. KEYS[1]
    return redis.call("ZCOUNT", key, 0, 0)
`)

var scriptAcctInfo = redis.NewScript(`
	local accts = redis.call("hgetall", "accounts:")
	local res = {}
	local index = 0
	for i, acc in ipairs(accts) do
        if i % 2 == 1 then
                local acct_key = "account:" .. acc;
                res[index] = {acc, redis.call('HGET', acct_key, 'bytes'), redis.call("HGET", acct_key, 'objects')}
                index = index + 1
        end;
	end;
	return cjson.encode(res);
`)

type client struct {
	Addr       string
	Client     *redis.Client
	IsAccount  bool
	IsCH       bool
	IsBucketDB bool
}

type redisDaemon struct {
	daemon.Daemon

	clients []*client
}

var whitelist = map[string]bool{
	"used_memory":                 true,
	"used_memory_rss":             true,
	"used_memory_lua":             true,
	"mem_fragmentation_ratio":     true,
	"rdb_changes_since_last_save": true,
	"total_connections_received":  true,
	"total_commands_processed":    true,
	"instantaneous_ops_per_sec":   true,
	"total_net_input_bytes":       true,
	"total_net_output_bytes":      true,
	"keyspace_hits":               true,
	"keyspace_misses":             true,
	"role":                        true,
	"connected_slaves":            true,
	"repl_backlog_size":           true,
	"db0":                         true,
}

var keysRegexp = regexp.MustCompile(`keys=(\d+)`)

func MakeDaemon() *redisDaemon {
	return &redisDaemon{
		clients: []*client{},
	}
}

func (d *redisDaemon) ID() string {
	return "Redis"
}

func (d *redisDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	d.reload(services)

	return nil
}

func (d *redisDaemon) OnServicesChange(services common.ServicesMap) error {
	log.Println("INFO", "redis module", "reloading config")

	d.reload(services)

	return nil
}

func (d *redisDaemon) reload(services common.ServicesMap) {
	d.cleanup()

	redisInstances, ok := services["redis"]
	if !ok || len(redisInstances) == 0 {
		return
	}

	clients := []*client{}
	var wg sync.WaitGroup
	var mutex sync.Mutex
	for _, svc := range redisInstances {
		wg.Add(1)
		go func(service *common.Service) {
			defer wg.Done()

			target := fmt.Sprintf("%s:%d", service.IP, service.Port)
			newClient := &client{
				Addr: target,
				Client: redis.NewClient(&redis.Options{
					Addr: target,
				}),
			}
			val, err := newClient.Client.HLen("accounts:").Result()
			// TODO: handle error
			if err == nil && val > 0 {
				newClient.IsAccount = true
			}

			mutex.Lock()
			log.Printf("new redis client: %s", newClient.Addr)
			clients = append(clients, newClient)
			mutex.Unlock()
		}(svc)
	}

	wg.Wait()

	d.clients = clients
}

func (d *redisDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup
	for _, c := range d.clients {
		if c.IsAccount {
			wg.Add(1)
			go func(c *client) {
				defer wg.Done()

				d.collectAccount(c, m)
			}(c)
		}
		wg.Add(1)
		go func(c *client) {
			defer wg.Done()

			d.collectStats(c, m)
		}(c)
	}

	wg.Wait()

	return nil
}

func (d *redisDaemon) cleanup() {
	for _, c := range d.clients {
		if err := c.Client.Close(); err != nil {
			log.Printf("WARN Failed to close redis client: %s", c.Addr)
		}
	}
	d.clients = []*client{}
}

func (d *redisDaemon) Close(ctx context.Context) error {
	d.cleanup()

	return nil
}

func (d *redisDaemon) collectStats(c *client, m *common.MetricData) {
	data, err := c.Client.Info().Result()
	if err != nil {
		m.IncrError("redis", "redis")
		log.Println("WARN", "when collecting redis stats", c, err)
		return
	}
	for _, line := range strings.Split(data, "\r\n") {
		kv := strings.Split(line, ":")
		if len(kv) != 2 {
			continue
		}
		if _, ok := whitelist[kv[0]]; ok {
			// Match keys in db entry
			if strings.HasPrefix(kv[0], "db") {
				keys := keysRegexp.FindStringSubmatch(kv[1])
				if len(keys) > 1 {
					valKeys, err := strconv.ParseFloat(keys[1], 64)
					if err != nil {
						m.IncrError("redis", "redis")
						log.Println("WARN", "when collecting redis stats", c, err)
						continue
					}
					m.Set(&common.Metric{
						Name: "redis_keys",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: valKeys,
					})
				}
				// Format role:master or role:slave
			} else if kv[0] == "role" {
				if kv[1] == "master" {
					m.Set(&common.Metric{
						Name: "redis_is_master",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: float64(1),
					})
				} else {
					m.Set(&common.Metric{
						Name: "redis_is_master",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: float64(0),
					})
					m.Set(&common.Metric{
						Name: "connected_slaves",
						Labels: [][]string{
							[]string{"addr", c.Addr},
						},
						Value: float64(0),
					})
				}
			} else {
				val, err := strconv.ParseFloat(kv[1], 64)
				if err != nil {
					m.IncrError("redis", "redis")
					log.Println("WARN", "when collecting redis stats", c, err)
					continue
				}
				m.Set(&common.Metric{
					Name: "redis_" + kv[0],
					Labels: [][]string{
						[]string{"addr", c.Addr},
					},
					Value: val,
				})
			}
		}
	}
}

func (d *redisDaemon) collectAccount(c *client, m *common.MetricData) {
	acctInfo, err := scriptAcctInfo.Run(c.Client, []string{}, 0).Result()
	if err != nil {
		log.Println("WARN", "when collecting account info", c, err)
		return
	}
	acctObj := map[string][]string{}
	err = json.Unmarshal([]byte(acctInfo.(string)), &acctObj)
	if err != nil {
		m.IncrError("redis", "account")
		log.Println("WARN", "when collecting account info", c, err)
		return
	}
	for _, data := range acctObj {
		valKiB, err := strconv.ParseFloat(data[1], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Println("WARN", "when collecting account objects", c, err)
			continue
		}
		valObj, err := strconv.ParseFloat(data[2], 64)
		if err != nil {
			m.IncrError("redis", "account")
			log.Println("WARN", "when collecting account objects", c, err)
			continue
		}
		m.Set(&common.Metric{
			Name: "account_kiB",
			Labels: [][]string{
				[]string{"account", data[0]},
			},
			Value: valKiB / 1024,
		})
		m.Set(&common.Metric{
			Name: "account_objects",
			Labels: [][]string{
				[]string{"account", data[0]},
			},
			Value: valObj,
		})
	}

	accounts, err := scriptGetAccounts.Run(c.Client, []string{}, 0).Result()
	if err != nil {
		m.IncrError("redis", "account")
		log.Println("WARN", "when listing accounts", c, err)
		return
	}
	for _, acct := range accounts.([]interface{}) {
		if acct == "1" {
			continue
		}
		count, err := scriptGetContCount.Run(c.Client, []string{acct.(string)}, 1).Result()
		if err != nil {
			m.IncrError("redis", "account")
			log.Println("WARN", "when collecting account containers", acct, err)
			continue
		}
		m.Set(&common.Metric{
			Name: "account_containers",
			Labels: [][]string{
				[]string{"addr", c.Addr},
				[]string{"account", acct.(string)},
			},
			Value: float64(count.(int64)),
		})
	}
}
