package tcp

import (
	"fmt"
	"log"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
)

type tcpDaemon struct {
	daemon.Daemon

	memcachedServices  []*common.Service
	beanstalkdServices []*common.Service
	zookeeperServices  []*common.Service
	timeout            time.Duration
}

func MakeDaemon() *tcpDaemon {
	return &tcpDaemon{
		timeout: 3 * time.Second, // TODO
	}
}

func (d *tcpDaemon) ID() string {
	return "TCP"
}

func (d *tcpDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	d.reload(services)

	return nil
}

func (d *tcpDaemon) OnServicesChange(services common.ServicesMap) error {
	d.reload(services)

	return nil
}

func (d *tcpDaemon) reload(services common.ServicesMap) {
	d.memcachedServices = services["memcached"]
	d.beanstalkdServices = services["beanstalkd"]
	d.zookeeperServices = services["zookeeper"]
}

func (d *tcpDaemon) Collect(m *common.MetricData) error {
	var wg sync.WaitGroup

	d.collect(d.collectMemcached, &wg, d.memcachedServices, m)
	d.collect(d.collectBeanstalkd, &wg, d.beanstalkdServices, m)
	d.collect(d.collectZookeeper, &wg, d.zookeeperServices, m)

	wg.Wait()

	return nil
}

func (d *tcpDaemon) collect(collectFunc func(*common.Service, *common.MetricData), wg *sync.WaitGroup, services []*common.Service, m *common.MetricData) {
	for _, svc := range services {
		wg.Add(1)
		go func(service *common.Service) {
			defer wg.Done()

			collectFunc(service, m)
		}(svc)
	}
}

func (d *tcpDaemon) collectMemcached(svc *common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	data, err := common.TCPCommand(target, "stats\r\n", " ", 1, d.timeout)
	if err != nil {
		m.IncrError("tcp", "memcached")
		log.Println("WARN", "in collecting memcached metrics", svc, err)
	}
	for _, v := range data {
		val, err := strconv.ParseFloat(v.Value.(string), 64)
		if err != nil {
			continue
		}
		m.Set(&common.Metric{
			Name: "memcached_" + v.Key,
			Labels: [][]string{
				[]string{"service", target},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
			},
			Value: val,
		})
	}
}

func (d *tcpDaemon) collectZookeeper(svc *common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	data, err := common.TCPCommand(target, "mntr\n", "\t", 0, d.timeout)
	if err != nil {
		m.IncrError("tcp", "zookeeper")
		log.Println("WARN", "in collecting zookeeper metrics", svc, err)
	}

	for _, v := range data {
		val, err := strconv.ParseFloat(v.Value.(string), 64)
		if err != nil {
			continue
		}
		m.Set(&common.Metric{
			Name: v.Key,
			Labels: [][]string{
				[]string{"service", target},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
			},
			Value: val,
		})
	}
}

func (d *tcpDaemon) collectBeanstalkd(svc *common.Service, m *common.MetricData) {
	target := fmt.Sprintf("%s:%d", svc.IP, svc.Port)

	data, err := common.TCPCommand(target, "stats\r\n", ": ", 0, d.timeout)
	if err != nil {
		m.IncrError("tcp", "beanstalkd")
		log.Println("WARN", "in collecting beanstalkd metrics", svc, err)
	}

	for _, v := range data {
		val, err := strconv.ParseFloat(v.Value.(string), 64)
		if err != nil {
			continue
		}
		m.Set(&common.Metric{
			Name: "beanstalkd_" + strings.Replace(v.Key, "-", "_", -1),
			Labels: [][]string{
				[]string{"service", target},
				[]string{"service_type", svc.Type},
				[]string{"addr", target},
			},
			Value: val,
		})
	}

	tubes, err := common.TCPCommand(target, "list-tubes\r\n", " ", 0, d.timeout)
	if err != nil {
		m.IncrError("tcp", "beanstalkd")
		log.Println("WARN", "in listing beanstalkd tubes", svc, err)
	}

	for _, tube := range tubes {
		data, err := common.TCPCommand(target, "stats-tube "+tube.Value.(string)+"\r\n", ": ", 0, d.timeout)
		if err != nil {
			m.IncrError("tcp", "beanstalkd")
			log.Println("WARN", "in collecting beanstalkd metrics for tube", tube.Value.(string), svc, err)
			continue
		}
		for _, v := range data {
			val, err := strconv.ParseFloat(v.Value.(string), 64)
			if err != nil {
				continue
			}
			m.Set(&common.Metric{
				Name: "beanstalkd_" + strings.Replace(v.Key, "-", "_", -1),
				Labels: [][]string{
					[]string{"service", target},
					[]string{"service_type", svc.Type},
					[]string{"addr", target},
					[]string{"tube", tube.Value.(string)},
				},
				Value: val,
			})
		}
	}
}
