package version

import (
	"encoding/base64"
	"errors"
	"fmt"
	"log"
	"os"
	"os/exec"
	"regexp"
	"strings"
	"time"

	"github.com/open-io/oio-exporter/common"
	"github.com/open-io/oio-exporter/daemon"
	"gopkg.in/yaml.v2"
)

type versionDaemon struct {
	daemon.Daemon

	packageCmd      string
	packager        string
	defaultVersions string
	versionPath     string

	versions []*version
}

type version struct {
	Name    string         `yaml:"name"`
	Type    string         `yaml:"type"`
	Package string         `yaml:"package"`
	Rpm     string         `yaml:"rpm"`
	Deb     string         `yaml:"deb"`
	Command string         `yaml:"command"`
	Extract string         `yaml:"extract"`
	Regex   *regexp.Regexp `yaml:"_disabled2"`
}

func parseVersions(path, defaultVersions string) ([]*version, error) {
	var versions struct {
		Versions []*version
	}

	if path == "" {
		data, err := base64.StdEncoding.DecodeString(defaultVersions)
		if err != nil {
			return nil, err
		}

		if err := yaml.Unmarshal(data, &versions); err != nil {
			return nil, err
		}
		log.Println("Versions definition loaded from defaults")
	} else {
		f, err := os.Open(path)
		if err != nil {
			return nil, err
		}
		defer f.Close()

		if err := yaml.NewDecoder(f).Decode(&versions); err != nil {
			return nil, err
		}
		log.Printf("Versions definition loaded from %s", path)
	}

	for _, version := range versions.Versions {
		if version.Name == "" {
			return nil, errors.New("name is missing in one of the version definition")
		}

		if version.Type == "package" {
			if version.Package == "" && (version.Deb == "" || version.Rpm == "") {
				return nil, fmt.Errorf("package type must have a package key or rpm and deb keys: %s", version.Name)
			}
		} else if version.Type == "command" {
			if version.Command == "" {
				return nil, fmt.Errorf("command type must have a command: %s", version.Name)
			}
			if version.Extract != "" {
				re, err := regexp.Compile(version.Extract)
				if err != nil {
					return nil, fmt.Errorf("invalid regex for %s: %w", version.Name, err)
				}
				version.Regex = re
			}
		} else {
			return nil, fmt.Errorf("unknown type %s for %s", version.Type, version.Name)
		}
	}

	return versions.Versions, nil
}

func MakeDaemon(versions, versionPath string) *versionDaemon {
	return &versionDaemon{
		defaultVersions: versions,
		versionPath:     versionPath,
	}
}

func (d *versionDaemon) ID() string {
	return "Versions"
}

func (d *versionDaemon) Interval() time.Duration {
	return 5 * time.Minute
}

func (d *versionDaemon) Init(services common.ServicesMap, targets []*common.Service) error {
	var packageCmd string
	var packager string
	if checkCMD("rpm") {
		packageCmd = "rpm -q --qf %{VERSION}-%{RELEASE}"
		packager = "rpm"
	} else if checkCMD("dpkg-query") {
		packageCmd = "dpkg-query -Wf=${Version}"
		packager = "deb"
	} else {
		log.Println("WARN", "in determining distro", "packages version will not be available")
		packageCmd = ""
		packager = ""
	}

	d.packageCmd = packageCmd
	d.packager = packager

	versions, err := parseVersions(d.versionPath, d.defaultVersions)
	if err != nil {
		return err
	}
	log.Printf("INFO: loaded %d versions", len(versions))
	d.versions = versions

	return nil
}

func (d *versionDaemon) Collect(m *common.MetricData) error {
	for _, version := range d.versions {
		var cmd string
		var regex *regexp.Regexp

		if version.Type == "package" {
			if d.packageCmd == "" {
				continue
			}

			cmd = d.packageCmd + " "
			if version.Package != "" {
				cmd += version.Package
			} else if d.packager == "rpm" && version.Rpm != "" {
				cmd += version.Rpm
			} else if d.packager == "deb" && version.Deb != "" {
				cmd += version.Deb
			} else {
				continue
			}
		} else if version.Type == "command" {
			cmd = version.Command
			regex = version.Regex
		}

		out, err := d.runCommand(cmd, regex)
		if err != nil {
			m.IncrError("command", version.Name)
			log.Println("WARN", "in running command", cmd, version.Name, err)
			continue
		}

		m.Set(&common.Metric{
			Name: "version",
			Labels: [][]string{
				[]string{"service_type", version.Name},
				[]string{"version", out},
			},
			Value: float64(time.Now().Unix()),
		})
	}

	return nil
}

func checkCMD(cmd string) bool {
	_, err := exec.LookPath(cmd)
	return err == nil
}

func (d *versionDaemon) runCommand(cmd string, regex *regexp.Regexp) (string, error) {
	cmds := strings.Split(cmd, " ")
	out, err := exec.Command(cmds[0], cmds[1:]...).CombinedOutput()

	if err != nil {
		return "", err
	}

	outFmt := strings.Trim(strings.TrimSuffix(string(out), "\n"), " ")

	if regex == nil {
		return outFmt, nil
	}

	matches := regex.FindStringSubmatch(outFmt)
	if len(matches) < 1 {
		return "", nil
	}
	if len(matches) == 1 {
		return matches[0], nil
	}

	return matches[1], nil
}
